package com.sun.org.apache.xml.internal.serializer;

import org.xml.sax.ContentHandler;
import java.util.StringTokenizer;
import java.util.List;
import java.util.ArrayList;
import org.xml.sax.Attributes;
import java.io.UnsupportedEncodingException;
import java.io.OutputStreamWriter;
import java.util.Iterator;
import java.util.Set;
import java.util.Enumeration;
import java.util.Properties;
import javax.xml.transform.ErrorListener;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import com.sun.org.apache.xml.internal.serializer.utils.Utils;
import java.io.Writer;
import com.sun.org.apache.xml.internal.serializer.utils.WrappedRuntimeException;
import org.w3c.dom.Node;
import java.io.IOException;
import org.xml.sax.SAXException;
import com.sun.org.apache.xalan.internal.utils.SecuritySupport;
import java.io.OutputStream;
import java.lang.reflect.Method;

public abstract class ToStream extends SerializerBase
{
    private static final String COMMENT_BEGIN = "<!--";
    private static final String COMMENT_END = "-->";
    protected ToStream.BoolStack m_disableOutputEscapingStates;
    EncodingInfo m_encodingInfo;
    Method m_canConvertMeth;
    boolean m_triedToGetConverter;
    Object m_charToByteConverter;
    protected ToStream.BoolStack m_preserves;
    protected boolean m_ispreserve;
    protected boolean m_isprevtext;
    protected int m_maxCharacter;
    protected char[] m_lineSep;
    protected boolean m_lineSepUse;
    protected int m_lineSepLen;
    protected CharInfo m_charInfo;
    boolean m_shouldFlush;
    protected boolean m_spaceBeforeClose;
    boolean m_startNewLine;
    protected boolean m_inDoctype;
    boolean m_isUTF8;
    protected boolean m_cdataStartCalled;
    private boolean m_expandDTDEntities;
    private char m_highSurrogate;
    private boolean m_escaping;
    OutputStream m_outputStream;
    private boolean m_writer_set_by_user;
    
    public ToStream() {
        this.m_disableOutputEscapingStates = new ToStream.BoolStack();
        this.m_encodingInfo = new EncodingInfo(null, null);
        this.m_triedToGetConverter = false;
        this.m_charToByteConverter = null;
        this.m_preserves = new ToStream.BoolStack();
        this.m_ispreserve = false;
        this.m_isprevtext = false;
        this.m_maxCharacter = Encodings.getLastPrintable();
        this.m_lineSep = SecuritySupport.getSystemProperty("line.separator").toCharArray();
        this.m_lineSepUse = true;
        this.m_lineSepLen = this.m_lineSep.length;
        this.m_shouldFlush = true;
        this.m_spaceBeforeClose = false;
        this.m_inDoctype = false;
        this.m_isUTF8 = false;
        this.m_cdataStartCalled = false;
        this.m_expandDTDEntities = true;
        this.m_highSurrogate = '\0';
        this.m_escaping = true;
    }
    
    protected void closeCDATA() throws SAXException {
        try {
            this.m_writer.write("]]>");
            this.m_cdataTagOpen = false;
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
    }
    
    @Override
    public void serialize(Node node) throws IOException {
        try {
            TreeWalker walker = new TreeWalker(this);
            walker.traverse(node);
        }
        catch (SAXException se) {
            throw new WrappedRuntimeException(se);
        }
    }
    
    static final boolean isUTF16Surrogate(char c) {
        return (c & 'ﰀ') == 0xD800;
    }
    
    protected final void flushWriter() throws SAXException {
        Writer writer = this.m_writer;
        if (null != writer) {
            try {
                if ((writer instanceof WriterToUTF8Buffered)) {
                    if (this.m_shouldFlush) {
                        ((WriterToUTF8Buffered)writer).flush();
                    }
                    else {
                        ((WriterToUTF8Buffered)writer).flushBuffer();
                    }
                }
                if ((writer instanceof WriterToASCI)) {
                    if (this.m_shouldFlush) {
                        writer.flush();
                    }
                }
                else {
                    writer.flush();
                }
            }
            catch (IOException ioe) {
                throw new SAXException(ioe);
            }
        }
    }
    
    @Override
    public OutputStream getOutputStream() {
        return this.m_outputStream;
    }
    
    @Override
    public void elementDecl(String name, String model) throws SAXException {
        if (this.m_inExternalDTD) {
            return;
        }
        try {
            Writer writer = this.m_writer;
            this.DTDprolog();
            writer.write("<!ELEMENT ");
            writer.write(name);
            writer.write(32);
            writer.write(model);
            writer.write(62);
            writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        return;
    }
    
    @Override
    public void internalEntityDecl(String name, String value) throws SAXException {
        if (this.m_inExternalDTD) {
            return;
        }
        try {
            this.DTDprolog();
            this.outputEntityDecl(name, value);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        return;
    }
    
    void outputEntityDecl(String name, String value) throws IOException {
        Writer writer = this.m_writer;
        writer.write("<!ENTITY ");
        writer.write(name);
        writer.write(" \"");
        writer.write(value);
        writer.write("\">");
        writer.write(this.m_lineSep, 0, this.m_lineSepLen);
    }
    
    protected final void outputLineSep() throws IOException {
        this.m_writer.write(this.m_lineSep, 0, this.m_lineSepLen);
    }
    
    @Override
    void setProp(String name, String val, boolean defaultVal) {
        if (val != null) {
            char first = SerializerBase.getFirstCharLocName(name);
            switch (first) {
                case 'c': {
                    if ("cdata-section-elements".equals(name)) {
                        this.addCdataSectionElements(val);
                        break;
                    }
                    break;
                }
                case 'd': {
                    if ("doctype-system".equals(name)) {
                        this.m_doctypeSystem = val;
                        break;
                    }
                    if ("doctype-public".equals(name)) {
                        this.m_doctypePublic = val;
                        if (val.startsWith("-//W3C//DTD XHTML")) {
                            this.m_spaceBeforeClose = true;
                            break;
                        }
                        break;
                    }
                    else {
                        break;
                    }
                }
                case 'e': {
                    String newEncoding = val;
                    if ("encoding".equals(name)) {
                        String possible_encoding = Encodings.getMimeEncoding(val);
                        if (possible_encoding != null) {
                            super.setProp("mime-name", possible_encoding, defaultVal);
                        }
                        String oldExplicitEncoding = this.getOutputPropertyNonDefault("encoding");
                        String oldDefaultEncoding = this.getOutputPropertyDefault("encoding");
                        if ((defaultVal && (oldDefaultEncoding == null || !oldDefaultEncoding.equalsIgnoreCase(newEncoding))) || (!defaultVal && (oldExplicitEncoding == null || !oldExplicitEncoding.equalsIgnoreCase(newEncoding)))) {
                            EncodingInfo encodingInfo = Encodings.getEncodingInfo(newEncoding);
                            if (newEncoding != null && encodingInfo.name == null) {
                                String msg = Utils.messages.createMessage("ER_ENCODING_NOT_SUPPORTED", new Object[] { newEncoding });
                                String msg2 = "Warning: encoding \"" + newEncoding + "\" not supported, using " + "UTF-8";
                                try {
                                    Transformer tran = super.getTransformer();
                                    if (tran != null) {
                                        ErrorListener errHandler = tran.getErrorListener();
                                        if (null != errHandler && this.m_sourceLocator != null) {
                                            errHandler.warning(new TransformerException(msg, this.m_sourceLocator));
                                            errHandler.warning(new TransformerException(msg2, this.m_sourceLocator));
                                        }
                                        else {
                                            System.out.println(msg);
                                            System.out.println(msg2);
                                        }
                                    }
                                    else {
                                        System.out.println(msg);
                                        System.out.println(msg2);
                                    }
                                }
                                catch (Exception ex) {}
                                newEncoding = "UTF-8";
                                val = "UTF-8";
                                encodingInfo = Encodings.getEncodingInfo(newEncoding);
                            }
                            if (!defaultVal || oldExplicitEncoding == null) {
                                this.m_encodingInfo = encodingInfo;
                                if (newEncoding != null) {
                                    this.m_isUTF8 = newEncoding.equals("UTF-8");
                                }
                                OutputStream os = this.getOutputStream();
                                if (os != null) {
                                    Writer w = this.getWriter();
                                    String oldEncoding = this.getOutputProperty("encoding");
                                    if ((w == null || !this.m_writer_set_by_user) && !newEncoding.equalsIgnoreCase(oldEncoding)) {
                                        super.setProp(name, val, defaultVal);
                                        this.setOutputStreamInternal(os, false);
                                    }
                                }
                            }
                        }
                        break;
                    }
                    else {
                        break;
                    }
                }
                case 'i': {
                    if ("{http://xml.apache.org/xalan}indent-amount".equals(name)) {
                        this.setIndentAmount(Integer.parseInt(val));
                        break;
                    }
                    if ("indent".equals(name)) {
                        boolean b = val.endsWith("yes");
                        this.m_doIndent = b;
                        break;
                    }
                    break;
                }
                case 'l': {
                    if ("{http://xml.apache.org/xalan}line-separator".equals(name)) {
                        this.m_lineSep = val.toCharArray();
                        this.m_lineSepLen = this.m_lineSep.length;
                        break;
                    }
                    break;
                }
                case 'm': {
                    if ("media-type".equals(name)) {
                        this.m_mediatype = val;
                        break;
                    }
                    break;
                }
                case 'o': {
                    if ("omit-xml-declaration".equals(name)) {
                        boolean b = val.endsWith("yes");
                        this.m_shouldNotWriteXMLHeader = b;
                        break;
                    }
                    break;
                }
                case 's': {
                    if ("standalone".equals(name)) {
                        if (defaultVal) {
                            this.setStandaloneInternal(val);
                            break;
                        }
                        this.m_standaloneWasSpecified = true;
                        this.setStandaloneInternal(val);
                        break;
                    }
                    else {
                        break;
                    }
                }
                case 'v': {
                    if ("version".equals(name)) {
                        this.m_version = val;
                        break;
                    }
                    break;
                }
            }
            super.setProp(name, val, defaultVal);
        }
    }
    
    @Override
    public void setOutputFormat(Properties format) {
        boolean shouldFlush = this.m_shouldFlush;
        if (format != null) {
            Enumeration<Object> propNames = format.propertyNames();
            while (propNames.hasMoreElements()) {
                String key = (String)propNames.nextElement();
                String value = format.getProperty(key);
                String explicitValue = (String)format.get(key);
                if (explicitValue == null && value != null) {
                    this.setOutputPropertyDefault(key, value);
                }
                if (explicitValue != null) {
                    this.setOutputProperty(key, explicitValue);
                }
                else {
                    continue;
                }
            }
        }
        String entitiesFileName = (String)format.get("{http://xml.apache.org/xalan}entities");
        if (null != entitiesFileName) {
            String method = (String)format.get("method");
            this.m_charInfo = CharInfo.getCharInfo(entitiesFileName, method);
        }
        this.m_shouldFlush = shouldFlush;
    }
    
    @Override
    public Properties getOutputFormat() {
        Properties def = new Properties();
        Set<String> s = this.getOutputPropDefaultKeys();
        for (String key : s) {
            String val = this.getOutputPropertyDefault(key);
            def.put(key, val);
        }
        Properties props = new Properties(def);
        Set<String> s2 = this.getOutputPropKeys();
        for (String key2 : s2) {
            String val2 = this.getOutputPropertyNonDefault(key2);
            if (val2 != null) {
                props.put(key2, val2);
            }
            else {
                continue;
            }
        }
        return props;
    }
    
    @Override
    public void setWriter(Writer writer) {
        this.setWriterInternal(writer, true);
    }
    
    private void setWriterInternal(Writer writer, boolean setByUser) {
        this.m_writer_set_by_user = setByUser;
        this.m_writer = writer;
        if (this.m_tracer != null) {
            boolean noTracerYet = true;
            Writer w2 = this.m_writer;
            while (w2 instanceof WriterChain) {
                if ((w2 instanceof SerializerTraceWriter)) {
                    noTracerYet = false;
                    break;
                }
                w2 = ((WriterChain)w2).getWriter();
            }
            if (noTracerYet) {
                this.m_writer = new SerializerTraceWriter(this.m_writer, this.m_tracer);
            }
        }
    }
    
    public boolean setLineSepUse(boolean use_sytem_line_break) {
        boolean oldValue = this.m_lineSepUse;
        this.m_lineSepUse = use_sytem_line_break;
        return oldValue;
    }
    
    @Override
    public void setOutputStream(OutputStream output) {
        this.setOutputStreamInternal(output, true);
    }
    
    private void setOutputStreamInternal(OutputStream output, boolean setByUser) {
        this.m_outputStream = output;
        String encoding = this.getOutputProperty("encoding");
        if ("UTF-8".equalsIgnoreCase(encoding)) {
            try {
                this.setWriterInternal(new WriterToUTF8Buffered(output), false);
            }
            catch (UnsupportedEncodingException e2) {
                e2.printStackTrace();
            }
        }
        else if ("WINDOWS-1250".equals(encoding) || "US-ASCII".equals(encoding) || "ASCII".equals(encoding)) {
            this.setWriterInternal(new WriterToASCI(output), false);
        }
        else if (encoding != null) {
            Writer osw = null;
            try {
                osw = Encodings.getWriter(output, encoding);
            }
            catch (UnsupportedEncodingException uee) {
                osw = null;
            }
            if (osw == null) {
                System.out.println("Warning: encoding \"" + encoding + "\" not supported, using " + "UTF-8");
                encoding = "UTF-8";
                this.setEncoding(encoding);
                try {
                    osw = Encodings.getWriter(output, encoding);
                }
                catch (UnsupportedEncodingException e) {
                    e.printStackTrace();
                }
            }
            this.setWriterInternal(osw, false);
        }
        else {
            Writer osw = new OutputStreamWriter(output);
            this.setWriterInternal(osw, false);
        }
    }
    
    @Override
    public boolean setEscaping(boolean escape) {
        boolean temp = this.m_escaping;
        this.m_escaping = escape;
        return temp;
    }
    
    protected void indent(int depth) throws IOException {
        if (this.m_startNewLine) {
            this.outputLineSep();
        }
        if (this.m_indentAmount > 0) {
            this.printSpace(depth * this.m_indentAmount);
        }
    }
    
    protected void indent() throws IOException {
        this.indent(this.m_elemContext.m_currentElemDepth);
    }
    
    private void printSpace(int n) throws IOException {
        Writer writer = this.m_writer;
        for (int i = 0; i < n; ++i) {
            writer.write(32);
        }
    }
    
    @Override
    public void attributeDecl(String eName, String aName, String type, String valueDefault, String value) throws SAXException {
        if (this.m_inExternalDTD) {
            return;
        }
        try {
            Writer writer = this.m_writer;
            this.DTDprolog();
            writer.write("<!ATTLIST ");
            writer.write(eName);
            writer.write(32);
            writer.write(aName);
            writer.write(32);
            writer.write(type);
            if (valueDefault != null) {
                writer.write(32);
                writer.write(valueDefault);
            }
            writer.write(62);
            writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        return;
    }
    
    @Override
    public Writer getWriter() {
        return this.m_writer;
    }
    
    @Override
    public void externalEntityDecl(String name, String publicId, String systemId) throws SAXException {
        try {
            this.DTDprolog();
            this.m_writer.write("<!ENTITY ");
            this.m_writer.write(name);
            if (publicId != null) {
                this.m_writer.write(" PUBLIC \"");
                this.m_writer.write(publicId);
            }
            else {
                this.m_writer.write(" SYSTEM \"");
                this.m_writer.write(systemId);
            }
            this.m_writer.write("\" >");
            this.m_writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    protected boolean escapingNotNeeded(char ch) {
        boolean ret;
        if (ch < '\u007f') {
            ret = (ch >= ' ' || '\n' == ch || '\r' == ch || '\t' == ch);
        }
        else {
            ret = this.m_encodingInfo.isInEncoding(ch);
        }
        return ret;
    }
    
    protected int writeUTF16Surrogate(char c, char[] ch, int i, int end) throws IOException, SAXException {
        int status = -1;
        if (i + 1 >= end) {
            this.m_highSurrogate = c;
            return status;
        }
        char high;
        char low;
        if (this.m_highSurrogate == '\0') {
            high = c;
            low = ch[i + 1];
            status = 0;
        }
        else {
            high = this.m_highSurrogate;
            low = c;
            this.m_highSurrogate = '\0';
        }
        if (!Encodings.isLowUTF16Surrogate(low)) {
            this.throwIOE(high, low);
        }
        Writer writer = this.m_writer;
        if (this.m_encodingInfo.isInEncoding(high, low)) {
            writer.write(new char[] { high, low }, 0, 2);
        }
        else {
            String encoding = this.getEncoding();
            if (encoding != null) {
                status = this.writeCharRef(writer, high, low);
            }
            else {
                writer.write(new char[] { high, low }, 0, 2);
            }
        }
        return status;
    }
    
    protected int accumDefaultEntity(Writer writer, char ch, int i, char[] chars, int len, boolean fromTextNode, boolean escLF) throws IOException {
        if (!escLF && '\n' == ch) {
            writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        else if ((fromTextNode && this.m_charInfo.isSpecialTextChar((int)ch)) || (!fromTextNode && this.m_charInfo.isSpecialAttrChar((int)ch))) {
            String outputStringForChar = this.m_charInfo.getOutputStringForChar(ch);
            if (null != outputStringForChar) {
                writer.write(outputStringForChar);
            }
            else {
                return i;
            }
        }
        else {
            return i;
        }
        return i + 1;
    }
    
    void writeNormalizedChars(char[] ch, int start, int length, boolean isCData, boolean useSystemLineSeparator) throws IOException, SAXException {
        Writer writer = this.m_writer;
        for (int end = start + length, i = start; i < end; ++i) {
            char c = ch[i];
            if ('\n' == c && useSystemLineSeparator) {
                writer.write(this.m_lineSep, 0, this.m_lineSepLen);
            }
            else if (isCData && !this.escapingNotNeeded(c)) {
                i = this.handleEscaping(writer, c, ch, i, end);
            }
            else if (isCData && i < end - 2 && ']' == c && ']' == ch[i + 1] && '>' == ch[i + 2]) {
                writer.write("]]]]><![CDATA[>");
                i += 2;
            }
            else if (this.escapingNotNeeded(c)) {
                if (isCData && !this.m_cdataTagOpen) {
                    writer.write("<![CDATA[");
                    this.m_cdataTagOpen = true;
                }
                writer.write((int)c);
            }
            else {
                i = this.handleEscaping(writer, c, ch, i, end);
            }
        }
    }
    
    private int handleEscaping(Writer writer, char c, char[] ch, int i, int end) throws IOException, SAXException {
        if (Encodings.isHighUTF16Surrogate(c) || Encodings.isLowUTF16Surrogate(c)) {
            if (this.writeUTF16Surrogate(c, ch, i, end) >= 0 && Encodings.isHighUTF16Surrogate(c)) {
                ++i;
            }
        }
        else {
            this.writeCharRef(writer, c);
        }
        return i;
    }
    
    public void endNonEscaping() throws SAXException {
        this.m_disableOutputEscapingStates.pop();
    }
    
    public void startNonEscaping() throws SAXException {
        this.m_disableOutputEscapingStates.push(true);
    }
    
    protected void cdata(char[] ch, int start, int length) throws SAXException {
        try {
            int old_start = start;
            if (this.m_elemContext.m_startTagOpen) {
                this.closeStartTag();
                this.m_elemContext.m_startTagOpen = false;
            }
            this.m_ispreserve = true;
            if (!this.m_cdataTagOpen && this.shouldIndent()) {
                this.indent();
            }
            boolean writeCDataBrackets = length >= 1 && this.escapingNotNeeded(ch[start]);
            if (writeCDataBrackets && !this.m_cdataTagOpen) {
                this.m_writer.write("<![CDATA[");
                this.m_cdataTagOpen = true;
            }
            if (this.isEscapingDisabled()) {
                this.charactersRaw(ch, start, length);
            }
            else {
                this.writeNormalizedChars(ch, start, length, true, this.m_lineSepUse);
            }
            if (writeCDataBrackets && ch[start + length - 1] == ']') {
                this.closeCDATA();
            }
            if (this.m_tracer != null) {
                super.fireCDATAEvent(ch, old_start, length);
            }
        }
        catch (IOException ioe) {
            throw new SAXException(Utils.messages.createMessage("ER_OIERROR", (Object[])null), ioe);
        }
    }
    
    private boolean isEscapingDisabled() {
        return this.m_disableOutputEscapingStates.peekOrFalse();
    }
    
    protected void charactersRaw(char[] ch, int start, int length) throws SAXException {
        if (this.m_inEntityRef) {
            return;
        }
        try {
            if (this.m_elemContext.m_startTagOpen) {
                this.closeStartTag();
                this.m_elemContext.m_startTagOpen = false;
            }
            this.m_ispreserve = true;
            this.m_writer.write(ch, start, length);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        return;
    }
    
    @Override
    public void characters(char[] chars, int start, int length) throws SAXException {
        if (length == 0 || (this.m_inEntityRef && !this.m_expandDTDEntities)) {
            return;
        }
        if (this.m_elemContext.m_startTagOpen) {
            this.closeStartTag();
            this.m_elemContext.m_startTagOpen = false;
        }
        else if (this.m_needToCallStartDocument) {
            this.startDocumentInternal();
        }
        if (this.m_cdataStartCalled || this.m_elemContext.m_isCdataSection) {
            this.cdata(chars, start, length);
            return;
        }
        if (this.m_cdataTagOpen) {
            this.closeCDATA();
        }
        if (this.m_disableOutputEscapingStates.peekOrFalse() || !this.m_escaping) {
            this.charactersRaw(chars, start, length);
            if (this.m_tracer != null) {
                super.fireCharEvent(chars, start, length);
            }
            return;
        }
        else {
            if (this.m_elemContext.m_startTagOpen) {
                this.closeStartTag();
                this.m_elemContext.m_startTagOpen = false;
            }
            try {
                int end = start + length;
                int lastDirty = start - 1;
                int i;
                char ch1;
                for (i = start; i < end && ((ch1 = chars[i]) == ' ' || (ch1 == '\n' && this.m_lineSepUse) || ch1 == '\r' || ch1 == '\t'); ++i) {
                    if (!this.m_charInfo.isTextASCIIClean((int)ch1)) {
                        lastDirty = (i = this.processDirty(chars, end, i, ch1, lastDirty, true));
                    }
                }
                if (i < end) {
                    this.m_ispreserve = true;
                }
                boolean isXML10 = "1.0".equals(this.getVersion());
                while (i < end) {
                    char ch2;
                    while (i < end && (ch2 = chars[i]) < '\u007f' && this.m_charInfo.isTextASCIIClean((int)ch2)) {
                        ++i;
                    }
                    if (i == end) {
                        break;
                    }
                    char ch3 = chars[i];
                    if ((isCharacterInC0orC1Range(ch3) || (!isXML10 && isNELorLSEPCharacter(ch3)) || !this.escapingNotNeeded(ch3) || this.m_charInfo.isSpecialTextChar((int)ch3)) && '\"' != ch3) {
                        lastDirty = (i = this.processDirty(chars, end, i, ch3, lastDirty, true));
                    }
                    ++i;
                }
                int startClean = lastDirty + 1;
                if (i > startClean) {
                    int lengthClean = i - startClean;
                    this.m_writer.write(chars, startClean, lengthClean);
                }
                this.m_isprevtext = true;
            }
            catch (IOException e) {
                throw new SAXException(e);
            }
            if (this.m_tracer != null) {
                super.fireCharEvent(chars, start, length);
            }
            return;
        }
    }
    
    private static boolean isCharacterInC0orC1Range(char ch) {
        return ch != '\t' && ch != '\n' && ch != '\r' && ((ch >= '\u007f' && ch <= '\u009f') || (ch >= '\u0001' && ch <= '\u001f'));
    }
    
    private static boolean isNELorLSEPCharacter(char ch) {
        return ch == '\u0085' || ch == '\u2028';
    }
    
    private int processDirty(char[] chars, int end, int i, char ch, int lastDirty, boolean fromTextNode) throws IOException, SAXException {
        int startClean = lastDirty + 1;
        if (i > startClean) {
            int lengthClean = i - startClean;
            this.m_writer.write(chars, startClean, lengthClean);
        }
        if ('\n' == ch && fromTextNode) {
            this.m_writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        else {
            startClean = this.accumDefaultEscape(this.m_writer, ch, i, chars, end, fromTextNode, false);
            i = startClean - 1;
        }
        return i;
    }
    
    @Override
    public void characters(String s) throws SAXException {
        if (this.m_inEntityRef && !this.m_expandDTDEntities) {
            return;
        }
        int length = s.length();
        if (length > this.m_charsBuff.length) {
            this.m_charsBuff = new char[length * 2 + 1];
        }
        s.getChars(0, length, this.m_charsBuff, 0);
        this.characters(this.m_charsBuff, 0, length);
        return;
    }
    
    protected int accumDefaultEscape(Writer writer, char ch, int i, char[] chars, int len, boolean fromTextNode, boolean escLF) throws IOException, SAXException {
        int pos = this.accumDefaultEntity(writer, ch, i, chars, len, fromTextNode, escLF);
        if (i == pos) {
            if (this.m_highSurrogate != '\0') {
                if (!Encodings.isLowUTF16Surrogate(ch)) {
                    this.throwIOE(this.m_highSurrogate, ch);
                }
                this.writeCharRef(writer, this.m_highSurrogate, ch);
                this.m_highSurrogate = '\0';
                return ++pos;
            }
            else if (Encodings.isHighUTF16Surrogate(ch)) {
                if (i + 1 >= len) {
                    this.m_highSurrogate = ch;
                    ++pos;
                }
                else {
                    char next = chars[++i];
                    if (!Encodings.isLowUTF16Surrogate(next)) {
                        this.throwIOE(ch, next);
                    }
                    this.writeCharRef(writer, ch, next);
                    pos += 2;
                }
            }
            else {
                if (isCharacterInC0orC1Range(ch) || ("1.1".equals(this.getVersion()) && isNELorLSEPCharacter(ch))) {
                    this.writeCharRef(writer, ch);
                }
                else if ((!this.escapingNotNeeded(ch) || (fromTextNode && this.m_charInfo.isSpecialTextChar((int)ch)) || (!fromTextNode && this.m_charInfo.isSpecialAttrChar((int)ch))) && this.m_elemContext.m_currentElemDepth > 0) {
                    this.writeCharRef(writer, ch);
                }
                else {
                    writer.write((int)ch);
                }
                ++pos;
            }
        }
        return pos;
    }
    
    private void writeCharRef(Writer writer, char c) throws IOException, SAXException {
        if (this.m_cdataTagOpen) {
            this.closeCDATA();
        }
        writer.write("&#");
        writer.write(Integer.toString((int)c));
        writer.write(59);
    }
    
    private int writeCharRef(Writer writer, char high, char low) throws IOException, SAXException {
        if (this.m_cdataTagOpen) {
            this.closeCDATA();
        }
        int codePoint = Encodings.toCodePoint(high, low);
        writer.write("&#");
        writer.write(Integer.toString(codePoint));
        writer.write(59);
        return codePoint;
    }
    
    private void throwIOE(char ch, char next) throws IOException {
        throw new IOException(Utils.messages.createMessage("ER_INVALID_UTF16_SURROGATE", new Object[] { Integer.toHexString((int)ch) + " " + Integer.toHexString((int)next) }));
    }
    
    @Override
    public void startElement(String namespaceURI, String localName, String name, Attributes atts) throws SAXException {
        if (this.m_inEntityRef) {
            return;
        }
        if (this.m_needToCallStartDocument) {
            this.startDocumentInternal();
            this.m_needToCallStartDocument = false;
        }
        else if (this.m_cdataTagOpen) {
            this.closeCDATA();
        }
        try {
            if (this.m_needToOutputDocTypeDecl && null != this.getDoctypeSystem()) {
                this.outputDocTypeDecl(name, true);
            }
            this.m_needToOutputDocTypeDecl = false;
            if (this.m_elemContext.m_startTagOpen) {
                this.closeStartTag();
                this.m_elemContext.m_startTagOpen = false;
            }
            if (namespaceURI != null) {
                this.ensurePrefixIsDeclared(namespaceURI, name);
            }
            this.m_ispreserve = false;
            if (this.shouldIndent() && this.m_startNewLine) {
                this.indent();
            }
            this.m_startNewLine = true;
            Writer writer = this.m_writer;
            writer.write(60);
            writer.write(name);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        if (atts != null) {
            this.addAttributes(atts);
        }
        this.m_elemContext = this.m_elemContext.push(namespaceURI, localName, name);
        this.m_isprevtext = false;
        if (this.m_tracer != null) {
            this.firePseudoAttributes();
        }
        return;
    }
    
    @Override
    public void startElement(String elementNamespaceURI, String elementLocalName, String elementName) throws SAXException {
        this.startElement(elementNamespaceURI, elementLocalName, elementName, null);
    }
    
    @Override
    public void startElement(String elementName) throws SAXException {
        this.startElement(null, null, elementName, null);
    }
    
    void outputDocTypeDecl(String name, boolean closeDecl) throws SAXException {
        if (this.m_cdataTagOpen) {
            this.closeCDATA();
        }
        try {
            Writer writer = this.m_writer;
            writer.write("<!DOCTYPE ");
            writer.write(name);
            String doctypePublic = this.getDoctypePublic();
            if (null != doctypePublic) {
                writer.write(" PUBLIC \"");
                writer.write(doctypePublic);
                writer.write(34);
            }
            String doctypeSystem = this.getDoctypeSystem();
            if (null != doctypeSystem) {
                if (null == doctypePublic) {
                    writer.write(" SYSTEM \"");
                }
                else {
                    writer.write(" \"");
                }
                writer.write(doctypeSystem);
                if (closeDecl) {
                    writer.write("\">");
                    writer.write(this.m_lineSep, 0, this.m_lineSepLen);
                    closeDecl = false;
                }
                else {
                    writer.write(34);
                }
            }
            boolean dothis = false;
            if (dothis && closeDecl) {
                writer.write(62);
                writer.write(this.m_lineSep, 0, this.m_lineSepLen);
            }
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
    }
    
    public void processAttributes(Writer writer, int nAttrs) throws IOException, SAXException {
        String encoding = this.getEncoding();
        for (int i = 0; i < nAttrs; ++i) {
            String name = this.m_attributes.getQName(i);
            String value = this.m_attributes.getValue(i);
            writer.write(32);
            writer.write(name);
            writer.write("=\"");
            this.writeAttrString(writer, value, encoding);
            writer.write(34);
        }
    }
    
    public void writeAttrString(Writer writer, String string, String encoding) throws IOException, SAXException {
        int len = string.length();
        if (len > this.m_attrBuff.length) {
            this.m_attrBuff = new char[len * 2 + 1];
        }
        string.getChars(0, len, this.m_attrBuff, 0);
        char[] stringChars = this.m_attrBuff;
        int i = 0;
        while (i < len) {
            char ch = stringChars[i];
            if (this.escapingNotNeeded(ch) && !this.m_charInfo.isSpecialAttrChar((int)ch)) {
                writer.write((int)ch);
                ++i;
            }
            else {
                i = this.accumDefaultEscape(writer, ch, i, stringChars, len, false, true);
            }
        }
    }
    
    @Override
    public void endElement(String namespaceURI, String localName, String name) throws SAXException {
        if (this.m_inEntityRef) {
            return;
        }
        this.m_prefixMap.popNamespaces(this.m_elemContext.m_currentElemDepth, null);
        try {
            Writer writer = this.m_writer;
            if (this.m_elemContext.m_startTagOpen) {
                if (this.m_tracer != null) {
                    super.fireStartElem(this.m_elemContext.m_elementName);
                }
                int nAttrs = this.m_attributes.getLength();
                if (nAttrs > 0) {
                    this.processAttributes(this.m_writer, nAttrs);
                    this.m_attributes.clear();
                }
                if (this.m_spaceBeforeClose) {
                    writer.write(" />");
                }
                else {
                    writer.write("/>");
                }
            }
            else {
                if (this.m_cdataTagOpen) {
                    this.closeCDATA();
                }
                if (this.shouldIndent()) {
                    this.indent(this.m_elemContext.m_currentElemDepth - 1);
                }
                writer.write(60);
                writer.write(47);
                writer.write(name);
                writer.write(62);
            }
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        if (!this.m_elemContext.m_startTagOpen && this.m_doIndent) {
            this.m_ispreserve = (!this.m_preserves.isEmpty() && this.m_preserves.pop());
        }
        this.m_isprevtext = false;
        if (this.m_tracer != null) {
            super.fireEndElem(name);
        }
        this.m_elemContext = this.m_elemContext.m_prev;
        return;
    }
    
    @Override
    public void endElement(String name) throws SAXException {
        this.endElement(null, null, name);
    }
    
    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        this.startPrefixMapping(prefix, uri, true);
    }
    
    @Override
    public boolean startPrefixMapping(String prefix, String uri, boolean shouldFlush) throws SAXException {
        int pushDepth;
        if (shouldFlush) {
            this.flushPending();
            pushDepth = this.m_elemContext.m_currentElemDepth + 1;
        }
        else {
            pushDepth = this.m_elemContext.m_currentElemDepth;
        }
        boolean pushed = this.m_prefixMap.pushNamespace(prefix, uri, pushDepth);
        if (pushed) {
            if ("".equals(prefix)) {
                String name = "xmlns";
                this.addAttributeAlways("http://www.w3.org/2000/xmlns/", name, name, "CDATA", uri, false);
            }
            else if (!"".equals(uri)) {
                String name = "xmlns:" + prefix;
                this.addAttributeAlways("http://www.w3.org/2000/xmlns/", prefix, name, "CDATA", uri, false);
            }
        }
        return pushed;
    }
    
    @Override
    public void comment(char[] ch, int start, int length) throws SAXException {
        int start_old = start;
        if (this.m_inEntityRef) {
            return;
        }
        if (this.m_elemContext.m_startTagOpen) {
            this.closeStartTag();
            this.m_elemContext.m_startTagOpen = false;
        }
        else if (this.m_needToCallStartDocument) {
            this.startDocumentInternal();
            this.m_needToCallStartDocument = false;
        }
        try {
            if (this.shouldIndent() && this.m_isStandalone) {
                this.indent();
            }
            int limit = start + length;
            boolean wasDash = false;
            if (this.m_cdataTagOpen) {
                this.closeCDATA();
            }
            if (this.shouldIndent() && !this.m_isStandalone) {
                this.indent();
            }
            Writer writer = this.m_writer;
            writer.write("<!--");
            for (int i = start; i < limit; wasDash = (ch[i] == '-'), ++i) {
                if (wasDash && ch[i] == '-') {
                    writer.write(ch, start, i - start);
                    writer.write(" -");
                    start = i + 1;
                }
            }
            if (length > 0) {
                int remainingChars = limit - start;
                if (remainingChars > 0) {
                    writer.write(ch, start, remainingChars);
                }
                if (ch[limit - 1] == '-') {
                    writer.write(32);
                }
            }
            writer.write("-->");
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
        this.m_startNewLine = true;
        if (this.m_tracer != null) {
            super.fireCommentEvent(ch, start_old, length);
        }
        return;
    }
    
    @Override
    public void endCDATA() throws SAXException {
        if (this.m_cdataTagOpen) {
            this.closeCDATA();
        }
        this.m_cdataStartCalled = false;
    }
    
    @Override
    public void endDTD() throws SAXException {
        try {
            if (this.m_needToCallStartDocument) {
                return;
            }
            if (this.m_needToOutputDocTypeDecl) {
                this.outputDocTypeDecl(this.m_elemContext.m_elementName, false);
                this.m_needToOutputDocTypeDecl = false;
            }
            Writer writer = this.m_writer;
            if (!this.m_inDoctype) {
                writer.write("]>");
            }
            else {
                writer.write(62);
            }
            writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        catch (IOException e) {
            throw new SAXException(e);
        }
    }
    
    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
    }
    
    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException {
        if (0 == length) {
            return;
        }
        this.characters(ch, start, length);
        return;
    }
    
    @Override
    public void skippedEntity(String name) throws SAXException {
    }
    
    @Override
    public void startCDATA() throws SAXException {
        this.m_cdataStartCalled = true;
    }
    
    @Override
    public void startEntity(String name) throws SAXException {
        if (name.equals("[dtd]")) {
            this.m_inExternalDTD = true;
        }
        if (!this.m_expandDTDEntities && !this.m_inExternalDTD) {
            this.startNonEscaping();
            this.characters("&" + name + 59);
            this.endNonEscaping();
        }
        this.m_inEntityRef = true;
    }
    
    protected void closeStartTag() throws SAXException {
        if (this.m_elemContext.m_startTagOpen) {
            try {
                if (this.m_tracer != null) {
                    super.fireStartElem(this.m_elemContext.m_elementName);
                }
                int nAttrs = this.m_attributes.getLength();
                if (nAttrs > 0) {
                    this.processAttributes(this.m_writer, nAttrs);
                    this.m_attributes.clear();
                }
                this.m_writer.write(62);
            }
            catch (IOException e) {
                throw new SAXException(e);
            }
            if (this.m_StringOfCDATASections != null) {
                this.m_elemContext.m_isCdataSection = this.isCdataSection();
            }
            if (this.m_doIndent) {
                this.m_isprevtext = false;
                this.m_preserves.push(this.m_ispreserve);
            }
        }
    }
    
    @Override
    public void startDTD(String name, String publicId, String systemId) throws SAXException {
        this.setDoctypeSystem(systemId);
        this.setDoctypePublic(publicId);
        this.m_elemContext.m_elementName = name;
        this.m_inDoctype = true;
    }
    
    @Override
    public int getIndentAmount() {
        return this.m_indentAmount;
    }
    
    @Override
    public void setIndentAmount(int m_indentAmount) {
        this.m_indentAmount = m_indentAmount;
    }
    
    protected boolean shouldIndent() {
        return this.m_doIndent && !this.m_ispreserve && !this.m_isprevtext && (this.m_elemContext.m_currentElemDepth > 0 || this.m_isStandalone);
    }
    
    private void setCdataSectionElements(String key, Properties props) {
        String s = props.getProperty(key);
        if (null != s) {
            ArrayList<String> al = new ArrayList();
            int l = s.length();
            boolean inCurly = false;
            StringBuilder buf = new StringBuilder();
            for (int i = 0; i < l; ++i) {
                char c = s.charAt(i);
                if (Character.isWhitespace(c)) {
                    if (!inCurly) {
                        if (buf.length() > 0) {
                            this.addCdataSectionElement(buf.toString(), al);
                            buf.setLength(0);
                            continue;
                        }
                        continue;
                    }
                }
                else if ('{' == c) {
                    inCurly = true;
                }
                else if ('}' == c) {
                    inCurly = false;
                }
                buf.append(c);
            }
            if (buf.length() > 0) {
                this.addCdataSectionElement(buf.toString(), al);
                buf.setLength(0);
            }
            this.setCdataSectionElements(al);
        }
    }
    
    private void addCdataSectionElement(String URI_and_localName, List<String> al) {
        StringTokenizer tokenizer = new StringTokenizer(URI_and_localName, "{}", false);
        String s1 = tokenizer.nextToken();
        String s2 = tokenizer.hasMoreTokens() ? tokenizer.nextToken() : null;
        if (null == s2) {
            al.add(null);
            al.add(s1);
        }
        else {
            al.add(s1);
            al.add(s2);
        }
    }
    
    @Override
    public void setCdataSectionElements(List<String> URI_and_localNames) {
        if (URI_and_localNames != null) {
            int len = URI_and_localNames.size() - 1;
            if (len > 0) {
                StringBuilder sb = new StringBuilder();
                for (int i = 0; i < len; i += 2) {
                    if (i != 0) {
                        sb.append(' ');
                    }
                    String uri = (String)URI_and_localNames.get(i);
                    String localName = (String)URI_and_localNames.get(i + 1);
                    if (uri != null) {
                        sb.append('{');
                        sb.append(uri);
                        sb.append('}');
                    }
                    sb.append(localName);
                }
                this.m_StringOfCDATASections = sb.toString();
            }
        }
        this.initCdataElems(this.m_StringOfCDATASections);
    }
    
    protected String ensureAttributesNamespaceIsDeclared(String ns, String localName, String rawName) throws SAXException {
        if (ns != null && ns.length() > 0) {
            int index = 0;
            String prefixFromRawName = ((index = rawName.indexOf(":")) < 0) ? "" : rawName.substring(0, index);
            if (index > 0) {
                String uri = this.m_prefixMap.lookupNamespace(prefixFromRawName);
                if (uri != null && uri.equals(ns)) {
                    return null;
                }
                this.startPrefixMapping(prefixFromRawName, ns, false);
                this.addAttribute("http://www.w3.org/2000/xmlns/", prefixFromRawName, "xmlns:" + prefixFromRawName, "CDATA", ns, false);
                return prefixFromRawName;
            }
            else {
                String prefix = this.m_prefixMap.lookupPrefix(ns);
                if (prefix == null) {
                    prefix = this.m_prefixMap.generateNextPrefix();
                    this.startPrefixMapping(prefix, ns, false);
                    this.addAttribute("http://www.w3.org/2000/xmlns/", prefix, "xmlns:" + prefix, "CDATA", ns, false);
                    return prefix;
                }
                return prefix;
            }
        }
        else {
            return null;
        }
    }
    
    void ensurePrefixIsDeclared(String ns, String rawName) throws SAXException {
        if (ns != null && ns.length() > 0) {
            int index;
            boolean no_prefix = (index = rawName.indexOf(":")) < 0;
            String prefix = no_prefix ? "" : rawName.substring(0, index);
            if (null != prefix) {
                String foundURI = this.m_prefixMap.lookupNamespace(prefix);
                if (null == foundURI || !foundURI.equals(ns)) {
                    this.startPrefixMapping(prefix, ns);
                    this.addAttributeAlways("http://www.w3.org/2000/xmlns/", no_prefix ? "xmlns" : prefix, no_prefix ? "xmlns" : ("xmlns:" + prefix), "CDATA", ns, false);
                }
            }
        }
    }
    
    @Override
    public void flushPending() throws SAXException {
        if (this.m_needToCallStartDocument) {
            this.startDocumentInternal();
            this.m_needToCallStartDocument = false;
        }
        if (this.m_elemContext.m_startTagOpen) {
            this.closeStartTag();
            this.m_elemContext.m_startTagOpen = false;
        }
        if (this.m_cdataTagOpen) {
            this.closeCDATA();
            this.m_cdataTagOpen = false;
        }
    }
    
    @Override
    public void setContentHandler(ContentHandler ch) {
    }
    
    @Override
    public boolean addAttributeAlways(String uri, String localName, String rawName, String type, String value, boolean xslAttribute) {
        int index = this.m_attributes.getIndex(rawName);
        boolean was_added;
        if (index >= 0) {
            String old_value = null;
            if (this.m_tracer != null) {
                old_value = this.m_attributes.getValue(index);
                if (value.equals(old_value)) {
                    old_value = null;
                }
            }
            this.m_attributes.setValue(index, value);
            was_added = false;
            if (old_value != null) {
                this.firePseudoAttributes();
            }
        }
        else {
            if (xslAttribute) {
                int colonIndex = rawName.indexOf(58);
                if (colonIndex > 0) {
                    String prefix = rawName.substring(0, colonIndex);
                    NamespaceMappings.MappingRecord existing_mapping = this.m_prefixMap.getMappingFromPrefix(prefix);
                    if (existing_mapping != null && existing_mapping.m_declarationDepth == this.m_elemContext.m_currentElemDepth && !existing_mapping.m_uri.equals(uri)) {
                        prefix = this.m_prefixMap.lookupPrefix(uri);
                        if (prefix == null) {
                            prefix = this.m_prefixMap.generateNextPrefix();
                        }
                        rawName = prefix + 58 + localName;
                    }
                }
                try {
                    this.ensureAttributesNamespaceIsDeclared(uri, localName, rawName);
                }
                catch (SAXException e) {
                    e.printStackTrace();
                }
            }
            this.m_attributes.addAttribute(uri, localName, rawName, type, value);
            was_added = true;
            if (this.m_tracer != null) {
                this.firePseudoAttributes();
            }
        }
        return was_added;
    }
    
    protected void firePseudoAttributes() {
        if (this.m_tracer != null) {
            try {
                this.m_writer.flush();
                StringBuffer sb = new StringBuffer();
                int nAttrs = this.m_attributes.getLength();
                if (nAttrs > 0) {
                    Writer writer = new ToStream.WritertoStringBuffer(sb);
                    this.processAttributes(writer, nAttrs);
                }
                sb.append('>');
                char[] ch = sb.toString().toCharArray();
                this.m_tracer.fireGenerateEvent(11, ch, 0, ch.length);
            }
            catch (IOException ex) {}
            catch (SAXException ex2) {}
        }
    }
    
    @Override
    public void setTransformer(Transformer transformer) {
        super.setTransformer(transformer);
        if (this.m_tracer != null && !(this.m_writer instanceof SerializerTraceWriter)) {
            this.m_writer = new SerializerTraceWriter(this.m_writer, this.m_tracer);
        }
    }
    
    @Override
    public boolean reset() {
        boolean wasReset = false;
        if (super.reset()) {
            this.resetToStream();
            wasReset = true;
        }
        return wasReset;
    }
    
    private void resetToStream() {
        this.m_cdataStartCalled = false;
        this.m_disableOutputEscapingStates.clear();
        this.m_escaping = true;
        this.m_inDoctype = false;
        this.m_ispreserve = false;
        this.m_ispreserve = false;
        this.m_isprevtext = false;
        this.m_isUTF8 = false;
        this.m_preserves.clear();
        this.m_shouldFlush = true;
        this.m_spaceBeforeClose = false;
        this.m_startNewLine = false;
        this.m_lineSepUse = true;
        this.m_expandDTDEntities = true;
    }
    
    @Override
    public void setEncoding(String encoding) {
        this.setOutputProperty("encoding", encoding);
    }
    
    @Override
    public void notationDecl(String name, String pubID, String sysID) throws SAXException {
        try {
            this.DTDprolog();
            this.m_writer.write("<!NOTATION ");
            this.m_writer.write(name);
            if (pubID != null) {
                this.m_writer.write(" PUBLIC \"");
                this.m_writer.write(pubID);
            }
            else {
                this.m_writer.write(" SYSTEM \"");
                this.m_writer.write(sysID);
            }
            this.m_writer.write("\" >");
            this.m_writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    @Override
    public void unparsedEntityDecl(String name, String pubID, String sysID, String notationName) throws SAXException {
        try {
            this.DTDprolog();
            this.m_writer.write("<!ENTITY ");
            this.m_writer.write(name);
            if (pubID != null) {
                this.m_writer.write(" PUBLIC \"");
                this.m_writer.write(pubID);
            }
            else {
                this.m_writer.write(" SYSTEM \"");
                this.m_writer.write(sysID);
            }
            this.m_writer.write("\" NDATA ");
            this.m_writer.write(notationName);
            this.m_writer.write(" >");
            this.m_writer.write(this.m_lineSep, 0, this.m_lineSepLen);
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    private void DTDprolog() throws SAXException, IOException {
        Writer writer = this.m_writer;
        if (this.m_needToOutputDocTypeDecl) {
            this.outputDocTypeDecl(this.m_elemContext.m_elementName, false);
            this.m_needToOutputDocTypeDecl = false;
        }
        if (this.m_inDoctype) {
            writer.write(" [");
            writer.write(this.m_lineSep, 0, this.m_lineSepLen);
            this.m_inDoctype = false;
        }
    }
    
    @Override
    public void setDTDEntityExpansion(boolean expand) {
        this.m_expandDTDEntities = expand;
    }
    
    public void addCdataSectionElements(String URI_and_localNames) {
        if (URI_and_localNames != null) {
            this.initCdataElems(URI_and_localNames);
        }
        if (this.m_StringOfCDATASections == null) {
            this.m_StringOfCDATASections = URI_and_localNames;
        }
        else {
            this.m_StringOfCDATASections = this.m_StringOfCDATASections + " " + URI_and_localNames;
        }
    }
    
    private class WritertoStringBuffer extends Writer
    {
        private final StringBuffer m_stringbuf;
        
        WritertoStringBuffer(StringBuffer sb) {
            this.m_stringbuf = sb;
        }
        
        @Override
        public void write(char[] arg0, int arg1, int arg2) throws IOException {
            this.m_stringbuf.append(arg0, arg1, arg2);
        }
        
        @Override
        public void flush() throws IOException {
        }
        
        @Override
        public void close() throws IOException {
        }
        
        @Override
        public void write(int i) {
            this.m_stringbuf.append((char)i);
        }
        
        @Override
        public void write(String s) {
            this.m_stringbuf.append(s);
        }
    }
    
    static final class BoolStack
    {
        private boolean[] m_values;
        private int m_allocatedSize;
        private int m_index;
        
        public BoolStack() {
            this(32);
        }
        
        public BoolStack(int size) {
            this.m_allocatedSize = size;
            this.m_values = new boolean[size];
            this.m_index = -1;
        }
        
        public final int size() {
            return this.m_index + 1;
        }
        
        public final void clear() {
            this.m_index = -1;
        }
        
        public final boolean push(boolean val) {
            if (this.m_index == this.m_allocatedSize - 1) {
                this.grow();
            }
            boolean[] values = this.m_values;
            int index = this.m_index + 1;
            this.m_index = index;
            return values[index] = val;
        }
        
        public final boolean pop() {
            return this.m_values[this.m_index--];
        }
        
        public final boolean popAndTop() {
            --this.m_index;
            return this.m_index >= 0 && this.m_values[this.m_index];
        }
        
        public final void setTop(boolean b) {
            this.m_values[this.m_index] = b;
        }
        
        public final boolean peek() {
            return this.m_values[this.m_index];
        }
        
        public final boolean peekOrFalse() {
            return this.m_index > -1 && this.m_values[this.m_index];
        }
        
        public final boolean peekOrTrue() {
            return this.m_index <= -1 || this.m_values[this.m_index];
        }
        
        public boolean isEmpty() {
            return this.m_index == -1;
        }
        
        private void grow() {
            this.m_allocatedSize *= 2;
            boolean[] newVector = new boolean[this.m_allocatedSize];
            System.arraycopy(this.m_values, 0, newVector, 0, this.m_index + 1);
            this.m_values = newVector;
        }
    }
}
